/*

RemindMe

*/

#include "date.hpp"

int IsLeapYear(unsigned int y)
{
    if (y % 4 == 0)
    {
        if (y % 100 == 0)
        {
            if (y % 400 == 0)
                return(1);
            return(0);
        }

        return(1);
    }

    return(0);
}

unsigned char DaysInMonth(unsigned char m, unsigned int y)
{
    switch (m)
    {
    case 1: return(31);
    case 2: return(IsLeapYear(y) ? 29 : 28);
    case 3: return(31);
    case 4: return(30);
    case 5: return(31);
    case 6: return(30);
    case 7: return(31);
    case 8: return(31);
    case 9: return(30);
    case 10: return(31);
    case 11: return(30);
    case 12: return(31);
    }

    return(0);
}

unsigned char DayOfWeek(unsigned char dom, unsigned char m, unsigned int y)
{
    unsigned int a = (14 - m) / 12;
    y -= a;
    m += 12 * a - 2;

    return(((dom + y + y/4 - y/100 + y/400 + (31*m)/12) % 7) + 1);
}

unsigned char NthDayOfMonth(unsigned char nth, unsigned char dow,
    unsigned char m, unsigned int y)
{
    unsigned char dom;
    unsigned char tdow;
    unsigned char dim = DaysInMonth(m, y);

    if (nth == NthLast)
    {
        dom = NthDayOfMonth(NthFifth, dow, m, y);
        if (dom == 0)
            dom = NthDayOfMonth(NthFourth, dow, m, y);
        return(dom);
    }

    for (dom = 1; dom < dim; dom++)
    {
        tdow = DayOfWeek(dom, m, y);
        if (tdow == dow)
        {
            nth -= 1;
            if (nth == 0)
                return(dom);
        }
    }

    return(0);
}

RMDateType::RMDateType(unsigned int g)
{
    Year = 1800;
    while (g >= (unsigned int) 365 + (IsLeapYear(Year) ? 1 : 0))
    {
        g -= 365;
        if (IsLeapYear(Year))
            g -= 1;
        Year += 1;
    }

    Month = 1;
    while (g >= DaysInMonth(Month, Year))
    {
        g -= DaysInMonth(Month, Year);
        Month += 1;
    }
    DayOfMonth = g + 1;
    DayOfWeek = ::DayOfWeek(DayOfMonth, Month, Year);
}

/*
Convert the date (day, month, and year) to the Gregorian date in days
since 1 January 1800.
*/
unsigned int RMDateType::Gregorian()
{
    unsigned int j = (Year - 1800) * 365;
    unsigned int y;

    for (y = 1800; y < Year; y++)
        if (IsLeapYear(y))
            j += 1;

    for (int m = 1; m < Month; m++)
        j += DaysInMonth(m, Year);
    j += DayOfMonth - 1;

    return(j);
}

/*
Convert the date (month and year) to the Gregorian date in months
since January 1800.
*/
unsigned int RMDateType::GregorianMonths()
{
    return((Year - 1800) * 12 + Month - 1);
}

void RMDateType::Yesterday()
{
    DayOfMonth -= 1;
    if (DayOfMonth == 0)
    {
        LastMonth();
        DayOfMonth = DaysInMonth(Month, Year);
    }
}

void RMDateType::Tomorrow()
{
    DayOfMonth += 1;
    if (DayOfMonth > DaysInMonth(Month, Year))
    {
        DayOfMonth = 1;
        NextMonth();
    }
}

void RMDateType::PlusDays(int d)
{
    while (d > 0)
    {
        Tomorrow();
        d -= 1;
    }
}

void RMDateType::MinusDays(int d)
{
    while (d > 0)
    {
        Yesterday();
        d -= 1;
    }
}

void RMDateType::NextMonth()
{
    Month += 1;
    if (Month == 13)
    {
        Month = 1;
        Year += 1;
    }
}

void RMDateType::LastMonth()
{
    Month -= 1;
    if (Month == 0)
    {
        Month = 12;
        Year -= 1;
    }
}

int RMDateType::Match(RMDateType dt)
{
    if ((Year == 0 || Year == dt.Year)
            && (Month == 0 || Month == dt.Month)
            && (DayOfMonth == 0 || DayOfMonth == dt.DayOfMonth)
            && (DayOfWeek == 0 || DayOfWeek == dt.DayOfWeek))
        return(1);
    return(0);
}

bool RMDateType::operator ==(RMDateType dt)
{
    if (DayOfMonth == dt.DayOfMonth && Month == dt.Month
            && Year == dt.Year)
        return(true);
    return(false);
}

