/*

RemindMe

*/

#include <windows.h>
#include <commctrl.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include "resource.hpp"
#include "date.hpp"
#include "app.hpp"

static HWND DialogHandle = 0;

static INT_PTR CALLBACK MainDialogProc(HWND hwnd, UINT message, WPARAM wparam,
    LPARAM lparam)
{
    switch (message)
    {
        case WM_INITDIALOG:
            DialogHandle = hwnd;
            EngineToday();
            EngineUpdate();
            return(1);

        case WM_COMMAND:
            if (LOWORD(wparam) == BUTTON_TODAY)
                EngineToday();
            else if (LOWORD(wparam) == BUTTON_PREVIOUS)
                EnginePrevious();
            else if (LOWORD(wparam) == BUTTON_NEXT)
                EngineNext();
            else if (LOWORD(wparam) == BUTTON_UPDATE)
                EngineUpdate();
            else
                break;
            return(1);

        case WM_SYSCOMMAND:
            if (wparam == SC_CLOSE)
            {
                PostQuitMessage(0);
                return(1);
            }
            break;
    }

    return(0);
}

void UISetDate(RMDateType dt)
{
    SYSTEMTIME st;

    SendMessage(GetDlgItem(DialogHandle, DATEPICKER_DATE), DTM_GETSYSTEMTIME,
            0, (LPARAM) &st);
    st.wDay = dt.DayOfMonth;
    st.wMonth = dt.Month;
    st.wYear = dt.Year;
    SendMessage(GetDlgItem(DialogHandle, DATEPICKER_DATE), DTM_SETSYSTEMTIME,
            GDT_VALID, (LPARAM) &st);
}

RMDateType UIGetDate()
{
    RMDateType dt;
    SYSTEMTIME st;

    SendMessage(GetDlgItem(DialogHandle, DATEPICKER_DATE), DTM_GETSYSTEMTIME,
            0, (LPARAM) &st);
    dt.DayOfMonth = (unsigned char) st.wDay;
    dt.Month = (unsigned char) st.wMonth;
    dt.Year = st.wYear;
    return(dt);
}

static const char * DayOfWeekNames[7] = {"Sunday", "Monday", "Tuesday",
    "Wednesday", "Thursday", "Friday", "Saturday"};

static const char * MonthNames[12] = {"January", "February", "March",
    "April", "May", "June", "July", "August", "September", "October",
    "November", "December"};

void UIUpdateDate(RMDateType dt)
{
    char buf[128];

    wsprintf(buf, "%s %d %s %d", DayOfWeekNames[DayOfWeek(dt.DayOfMonth,
            dt.Month, dt.Year) - 1], dt.DayOfMonth, MonthNames[dt.Month - 1],
            dt.Year);
    SetWindowText(GetDlgItem(DialogHandle, STATIC_DATE), buf);
}

int UIGetBefore()
{
    return(7);
}

int UIGetAfter()
{
    return(14);
}

static void UIClearText()
{
    SendMessage(GetDlgItem(DialogHandle, EDIT_TEXT), LB_RESETCONTENT, 0, 0);
}

static void UIAppendText(char * s)
{
    SendMessage(GetDlgItem(DialogHandle, EDIT_TEXT), LB_ADDSTRING, 0,
            (LPARAM) s);
}

static const char * ShortMonthNames[12] = {"jan", "feb", "mar",
    "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"};

static char RunRMEPath[MAX_PATH];

static void RunErrorMessage()
{
    UIClearText();
    UIAppendText("Error executing runrme");
    UIAppendText(RunRMEPath);
}

static char RunBuffer[1024 * 32];

void UIRunRME(RMDateType dt, int b, int a)
{
    HANDLE ih, oh;
    SECURITY_ATTRIBUTES sa;
    PROCESS_INFORMATION pi;
    STARTUPINFO si;
    DWORD ra, ls, le;
    BOOL ret;

    memset(&sa, 0, sizeof(sa));
    sa.nLength = sizeof(sa);
    sa.bInheritHandle = TRUE;
    sa.lpSecurityDescriptor = 0;

    memset(&si, 0, sizeof(si));
    si.cb = sizeof(si);
    si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
    si.wShowWindow = SW_HIDE;

    if (CreatePipe(&ih, &si.hStdOutput, &sa, 0) == 0
            || CreatePipe(&si.hStdInput, &oh, &sa, 0) == 0)
    {
        RunErrorMessage();
        return;
    }

    si.hStdError = si.hStdOutput;

    sprintf(RunBuffer, "%s -b %d -a %d %d %s %d", RunRMEPath, b, a,
        dt.DayOfMonth, ShortMonthNames[dt.Month - 1], dt.Year);

    ret = CreateProcess(0, RunBuffer, 0, 0, TRUE, CREATE_NEW_CONSOLE, 0, 0,
            &si, &pi);
    CloseHandle(si.hStdInput);
    CloseHandle(si.hStdOutput);
    if (ret == 0)
    {
        CloseHandle(ih);
        CloseHandle(oh);
        RunErrorMessage();
        return;
    }

    CloseHandle(pi.hThread);
    CloseHandle(pi.hProcess);

    if (ReadFile(ih, RunBuffer, sizeof(RunBuffer), &ra, 0) == 0)
        RunErrorMessage();
    else
    {
        UIClearText();
        RunBuffer[ra] = 0;

        for (ls = 0, le = 0; ls < ra;)
        {
            while (le < ra && RunBuffer[le] != '\r')
                le += 1;
            RunBuffer[le] = 0;
            le += 2;
            UIAppendText(RunBuffer + ls);
            ls = le;
        }
    }

    CloseHandle(ih);
    CloseHandle(oh);
}

int WINAPI WinMain(HINSTANCE hi, HINSTANCE hip, LPSTR cl, int sf)
{
    INITCOMMONCONTROLSEX icc;
    char mfn[MAX_PATH];
    char * fp;

    icc.dwSize = sizeof(icc);
    icc.dwICC = ICC_DATE_CLASSES;
    InitCommonControlsEx(&icc);

    GetModuleFileName(hi, mfn, MAX_PATH);
    if (GetFullPathName(mfn, MAX_PATH, RunRMEPath, &fp) > MAX_PATH)
        RunRMEPath[0] = 0;
    else
        *fp = 0;
    strcat(RunRMEPath, "runrme.exe");

    DialogBox(hi, MAKEINTRESOURCE(DIALOG_MAIN), 0, MainDialogProc);
    return(0);
}

/*
            GetWindowRect(DialogHandle, &rect);
            SystemParametersInfo(SPI_GETWORKAREA, 0, &wa, 0);
            w = rect.right - rect.left + GetSystemMetrics(SM_CXFIXEDFRAME) * 2;
            h = rect.bottom - rect.top + GetSystemMetrics(SM_CYFIXEDFRAME) * 2
                    + GetSystemMetrics(SM_CYCAPTION);
            MoveWindow(hwnd, (wa.right - w) / 2, (wa.bottom - h) / 2, w, h, 0);
            return(0);
*/

