/*

(Yet Another) Tiny Scheme

*/

#include "tscm.hpp"

int ChAlphabeticP(TSCh ch)
{
    if (ch >= 'a' && ch <= 'z' || ch >= 'A' && ch <= 'Z')
        return(1);
    return(0);
}

int ChNumericP(TSCh ch)
{
    if (ch >= '0' && ch <= '9')
        return(1);
    return(0);
}

int ChWhitespaceP(TSCh ch)
{
    // space, tab, line feed, form feed, carriage return
    if (ch == 32 || ch == 9 || ch == 10 || ch == 12 || ch == 13)
        return(1);
    return(0);
}

TSCh ChUpCase(TSCh ch)
{
    if (ch >= 'a' && ch <= 'z')
        return(ch - 'a' + 'A');
    return(ch);
}

TSCh ChDownCase(TSCh ch)
{
    if (ch >= 'A' && ch <= 'Z')
        return(ch - 'A' + 'a');
    return(ch);
}

int ChCompare(TSCh ch1, TSCh ch2)
{
    return(ch1 - ch2);
}

int ChCiCompare(TSCh ch1, TSCh ch2)
{
    return(ChDownCase(ch1) - ChDownCase(ch2));
}

static int StringLength(TSCh * s)
{
    int sl = 0;

    while (*s != 0)
    {
        s += 1;
        sl += 1;
    }

    return(sl);
}

TSCh * StringDup(int sl, TSCh * s)
{
    TSCh * ns = new TSCh[sl + 1];
    int sdx;

    for (sdx = 0; sdx < sl; sdx++)
        ns[sdx] = s[sdx];

    ns[sl] = 0;
    return(ns);
}

int StringEqual(TSCh * s1, TSCh * s2, int sl)
{
    int idx;

    for (idx = 0; idx < sl; idx++)
        if (s1[idx] != s2[idx])
            return(0);

    return(1);
}

const static char Digits[] = {"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"};

int NumberAsString(TSFixnum n, TSCh * s, TSFixnum b)
{
    int sl = 0;

    if (n < 0)
    {
        s[sl] = '-';
        sl += 1;
        n *= -1;
    }

    if (n >= b)
    {
        sl += NumberAsString(n / b, s + sl, b);
        s[sl] = Digits[n % b];
        sl += 1;
    }
    else
    {
        s[sl] = Digits[n];
        sl += 1;
    }

    return(sl);
}

int StringAsNumber(int sl, TSCh * s, TSFixnum * np)
{
    TSFixnum ns;
    TSFixnum n;
    int sdx;

    ns = 1;
    sdx = 0;
    if (s[0] == '-')
    {
        ns = -1;
        if (sl == 1)
            return(0);
        sdx += 1;
    }
    else if (s[0] == '+')
    {
        if (sl == 1)
            return(0);
        sdx += 1;
    }

    for (n = 0; sdx < sl; sdx++)
    {
        if (s[sdx] >= '0' && s[sdx] <= '9')
            n = n * 10 + s[sdx] - '0';
        else
            return(0);
    }

    *np = n * ns;
    return(1);
}

class TSBoolean : public TSObject
{
private:

    friend class TSWorld;

    TSBoolean(TSWorld * wld) : TSObject(wld, BooleanType) { /* Nothing. */ }
};

#define BooleanP(obj) (HeapP((obj)) && ((obj)->Type == BooleanType))

class TSEvaluateContext;
class TSContext : public TSObject
{
public:

    virtual TSObject * Evaluate(TSWorld * wld) = 0;
    virtual TSObject * GetValue(TSWorld * wld, TSSymbol * sym) = 0;
    virtual int SetValue(TSWorld * wld, TSSymbol * sym, TSObject * val) = 0;
    virtual TSLineNumberMap * GetLineNumberMap() = 0;
    virtual TSProcedure * GetProcedure() = 0;
    virtual TSEvaluateContext * AsEvaluateContext();

    TSContext * Stack;
    TSObject * Expression;

protected:

    TSContext(TSWorld * wld, TSObject * expr);
    virtual ~TSContext();
};

#define ContextP(obj) (HeapP((obj)) && ((obj)->Type == ContextType))

TSObject * Evaluate(TSWorld * wld, TSObject * obj)
{
    do
    {
// GC: need to protect obj

        if (HeapP(obj) == 0)
            break;
        obj = obj->Evaluate(wld);
    }
    while (ContextP(obj));

    return(obj);
}

static TSObject * EvaluateBlock(TSWorld * wld, TSObject * obj)
{
    wld->MustBe(PairType, obj);

    for (;;)
    {
        if (Rest(obj) == wld->EmptyList)
        {
            if (HeapP(First(obj)) == 0)
                return(First(obj));
            return(First(obj)->Evaluate(wld));
        }

        Evaluate(wld, First(obj));
        wld->MustBe(PairType, Rest(obj));
        obj = Rest(obj);
    }

    return(Undefined);
}

TSObject * TSObject::Call(TSWorld * wld, int argc, TSObject * argv[])
{
    wld->ThrowError("Not callable", this);
    return(Undefined);
}

TSObject * TSObject::Evaluate(TSWorld * wld)
{
    return(this);
}

TSObject * TSObject::Syntax(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    wld->ThrowError("Internal error", this);
    return(Undefined);
}

TSObject::TSObject(TSWorld * wld, TSObjectType ot)
{
    Type = ot;
    Next = wld->Objects;
    wld->Objects = this;
}

TSObject::~TSObject()
{
    // Nothing.
}

TSString::TSString(TSWorld * wld, int sl, TSCh * s) : TSObject(wld, StringType)
{
    Length = sl;
    if (s == 0)
        String = new TSCh[sl + 1];
    else
        String = StringDup(sl, s);
}

TSString::~TSString()
{
    delete String;
}

TSObject * TSVector::Evaluate(TSWorld * wld)
{
    wld->ThrowError("Vectors must be quoted", this);
    return(this);
}

TSVector::TSVector(TSWorld * wld, int vl, TSObject * obj)
    : TSObject(wld, VectorType)
{
    int vdx;

    Length = vl;
    Vector = new TSObject * [Length];
    for (vdx = 0; vdx < Length; vdx++)
        Vector[vdx] = obj;
}

TSVector::~TSVector()
{
    delete Vector;
}

TSObject * TSSymbol::Evaluate(TSWorld * wld)
{
    TSObject * obj;

    if (wld->Stack != 0)
    {
        obj = wld->Stack->GetValue(wld, this);
        if (obj != Undefined)
            return(obj);
    }

    if (Value == Undefined)
        wld->ThrowError("Undefined symbol", this);

    return(Value);
}

void TSSymbol::SetValue(TSWorld * wld, TSObject * val)
{
    if (wld->Stack->SetValue(wld, this, val) == 0)
        Value = val;
}

TSSymbol::TSSymbol(TSWorld * wld, TSString * name, TSSymbol * sym)
    : TSObject(wld, SymbolType)
{
    Name = name;
    Value = Undefined;
    Next = sym;
}

TSObject * TSPair::Evaluate(TSWorld * wld)
{
// GC: need to protect obj

    wld->Stack->Expression = this;
    TSObject * obj = ::Evaluate(wld, First);

    if (SyntaxP(obj))
    {
        wld->Stack->Expression = this;
        return(obj->Syntax(wld, Rest, this));
    }

    int argc = 0;
    TSObject * argv[128];
    TSObject * args = Rest;

    for (;;)
    {
        if (args == wld->EmptyList)
            break;
        wld->MustBe(PairType, args);

// GC: need to protect args

        wld->Stack->Expression = this;
        argv[argc] = ::Evaluate(wld, First(args));

        args = Rest(args);

        argc += 1;
        if (argc == sizeof(argv) / sizeof(TSObject *))
            wld->ThrowError("Too many arguments", this);
    }

    wld->Stack->Expression = this;
    return(obj->Call(wld, argc, argv));
}

TSPair::TSPair(TSWorld * wld, TSObject * first, TSObject * rest)
    : TSObject(wld, PairType)
{
    First = first;
    Rest = rest;
}

TSBuiltinInfo::TSBuiltinInfo(char * name, TSBuiltinFn fn, char * sig,
    char * filename, int ln)
{
    Name = name;
    Builtin = fn;
    Signature = sig;
    Filename = filename;
    LineNumber = ln;

    Next = BuiltinInfoList;
    BuiltinInfoList = this;
}

static void CheckType(TSWorld * wld, char sig, TSObject * obj)
{
    switch (sig)
    {
    case '.': break; // any type
    case 'n': wld->MustBe(NumberType, obj); break;
    case 's': wld->MustBe(StringType, obj); break;
    case 'p': wld->MustBe(PairType, obj); break;
    case 'S': wld->MustBe(SymbolType, obj); break;
    case 'c': wld->MustBe(CharacterType, obj); break;
    case 'v': wld->MustBe(VectorType, obj); break;
    }
}

TSObject * TSBuiltin::Call(TSWorld * wld, int argc, TSObject * argv[])
{
    char * sig = Info->Signature;
    int adx = 0;

    for (;;)
    {
        if (*sig == 0)
        {
            if (adx != argc)
                wld->ThrowError("Wrong number of arguments", this);
            break;
        }

        if (adx == argc && *sig != '*')
            wld->ThrowError("Wrong number of arguments", this);

        if (*sig == '*')
        {
            sig += 1;
            while (adx < argc)
            {
                CheckType(wld, *sig, argv[adx]);
                adx += 1;
            }
        }
        else
        {
            CheckType(wld, *sig, argv[adx]);
            adx += 1;
        }

        sig += 1;
    }

    return(Info->Builtin(wld, argc, argv));
}

TSBuiltin::TSBuiltin(TSWorld * wld, TSBuiltinInfo * bi)
    : TSObject(wld, BuiltinType)
{
    Info = bi;
}

static int ListLength(TSWorld * wld, TSObject * obj)
{
    int ll = 0;

    while (obj != wld->EmptyList)
    {
        wld->MustBe(PairType, obj);
        ll += 1;
        obj = Rest(obj);
    }

    return(ll);
}

static TSObject * ListToVector(TSWorld * wld, TSObject * obj)
{
    int vl = ListLength(wld, obj);
    TSObject * v = wld->MakeVector(vl, wld->Unspecified);
    int vdx;

    for (vdx = 0; vdx < vl; vdx++)
    {
        AsVector(v)->Vector[vdx] = First(obj);
        obj = Rest(obj);
    }

    return(v);
}

//
// Equivalence Predicates
//

static int EqvP(TSObject * obj1, TSObject * obj2)
{
    if (obj1 == obj2)
        return(1);
    if (ObjectType(obj1) == ObjectType(obj2))
    {
        if (NumberP(obj1) && AsFixnum(obj1) == AsFixnum(obj2))
            return(1);

        if (CharacterP(obj1) && AsCharacter(obj1) == AsCharacter(obj2))
            return(1);
    }

    return(0);
}

DefineBuiltin("eqv?", EqvP, "..")(TSWorld * wld, int argc, TSObject * argv[])
{
    if (EqvP(argv[0], argv[1]))
        return(wld->True);

    return(wld->False);
}

static int EqP(TSObject * obj1, TSObject * obj2)
{
    if (obj1 == obj2)
        return(1);

    return(0);
}

DefineBuiltin("eq?", EqP, "..")(TSWorld * wld, int argc, TSObject * argv[])
{
    if (EqP(argv[0], argv[1]))
        return(wld->True);

    return(wld->False);
}

static int EqualP(TSObject * obj1, TSObject * obj2)
{
    if (obj1 == obj2)
        return(1);

    if (ObjectType(obj1) != ObjectType(obj2))
        return(0);

    switch (ObjectType(obj1))
    {
        case PairType:
            if (EqualP(First(obj1), First(obj2)) == 0)
                return(0);
            return(EqualP(Rest(obj1), Rest(obj2)));
            break;

        case VectorType:
            if (AsVector(obj1)->Length != AsVector(obj2)->Length)
                return(0);
            for (int idx = 0; idx < AsVector(obj1)->Length; idx++)
                if (EqualP(AsVector(obj1)->Vector[idx],
                        AsVector(obj2)->Vector[idx]) == 0)
                    return(0);
            return(1);

        case StringType:
            if (AsString(obj1)->Length != AsString(obj2)->Length)
                return(0);
            for (int idx = 0; idx < AsString(obj1)->Length; idx++)
                if (AsString(obj1)->String[idx] !=
                        AsString(obj2)->String[idx])
                    return(0);
            return(1);

        case NumberType:
            if (AsFixnum(obj1) == AsFixnum(obj2))
                return(1);
            break;

        case CharacterType:
            if (AsCharacter(obj1) == AsCharacter(obj2))
                return(1);
            break;

        case SymbolType:
        case BooleanType:
        case UnspecifiedType:
        case EmptyListType:
        case EndOfFileType:
        case ProcedureType:
        case ContextType:
        case BuiltinType:
        case SyntaxType:
            break;
    }

    return(0);
}

DefineBuiltin("equal?", EqualP, "..")(TSWorld * wld, int argc, TSObject * argv[])
{
    if (EqualP(argv[0], argv[1]))
        return(wld->True);

    return(wld->False);
}

//
// Numbers
//

DefineBuiltin("number?", NumberP, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (NumberP(argv[0]))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("integer?", IntegerP, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (NumberP(argv[0]))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("=", NumberEqual, "nn*n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    int adx;

    for (adx = 1; adx < argc; adx++)
        if (AsFixnum(argv[0]) != AsFixnum(argv[adx]))
            return(wld->False);

    return(wld->True);
}

DefineBuiltin("<", NumberLess, "nn*n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    int adx = 0;

    do
    {
        adx += 1;
        if (adx == argc)
            return(wld->True);
    }
    while (AsFixnum(argv[0]) < AsFixnum(argv[adx]));

    return(wld->False);
}

DefineBuiltin(">", NumberGreater, "nn*n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    int adx = 0;

    do
    {
        adx += 1;
        if (adx == argc)
            return(wld->True);
    }
    while (AsFixnum(argv[0]) > AsFixnum(argv[adx]));

    return(wld->False);
}

DefineBuiltin("<=", NumberLessEqual, "nn*n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    int adx = 0;

    do
    {
        adx += 1;
        if (adx == argc)
            return(wld->True);
    }
    while (AsFixnum(argv[0]) <= AsFixnum(argv[adx]));

    return(wld->False);
}

DefineBuiltin(">=", NumberGreaterEqual, "nn*n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    int adx = 0;

    do
    {
        adx += 1;
        if (adx == argc)
            return(wld->True);
    }
    while (AsFixnum(argv[0]) >= AsFixnum(argv[adx]));

    return(wld->False);
}

DefineBuiltin("zero?", ZeroP, "n")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(AsFixnum(argv[0]) == 0 ? wld->True : wld->False);
}

DefineBuiltin("positive?", PositiveP, "n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(AsFixnum(argv[0]) > 0 ? wld->True : wld->False);
}

DefineBuiltin("negative?", NegativeP, "n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(AsFixnum(argv[0]) < 0 ? wld->True : wld->False);
}

DefineBuiltin("odd?", OddP, "n")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(AsFixnum(argv[0]) % 2 == 1 ? wld->True : wld->False);
}

DefineBuiltin("even?", EvenP, "n")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(AsFixnum(argv[0]) % 2 == 0 ? wld->True : wld->False);
}

DefineBuiltin("max", Max, "n*n")(TSWorld * wld, int argc, TSObject * argv[])
{
    TSFixnum n = AsFixnum(argv[0]);

    for (int adx = 1; adx < argc; adx++)
        if (AsFixnum(argv[adx]) > n)
            n = AsFixnum(argv[adx]);

    return(wld->MakeNumber(n));
}

DefineBuiltin("min", Min, "n*n")(TSWorld * wld, int argc, TSObject * argv[])
{
    TSFixnum n = AsFixnum(argv[0]);

    for (int adx = 1; adx < argc; adx++)
        if (AsFixnum(argv[adx]) < n)
            n = AsFixnum(argv[adx]);

    return(wld->MakeNumber(n));
}

DefineBuiltin("+", Plus, "*n")(TSWorld * wld, int argc, TSObject * argv[])
{
    int adx;
    TSFixnum n = 0;

    for (adx = 0; adx < argc; adx++)
        n += AsFixnum(argv[adx]);

    return(wld->MakeNumber(n));
}

DefineBuiltin("*", Times, "*n")(TSWorld * wld, int argc, TSObject * argv[])
{
    int adx;
    TSFixnum n = 1;

    for (adx = 0; adx < argc; adx++)
        n *= AsFixnum(argv[adx]);

    return(wld->MakeNumber(n));
}

DefineBuiltin("-", Subtract, "n*n")(TSWorld * wld, int argc, TSObject * argv[])
{
    int adx;
    TSFixnum n = AsFixnum(argv[0]);

    if (argc == 1)
        return(wld->MakeNumber(- n));

    for (adx = 1; adx < argc; adx++)
        n -= AsFixnum(argv[adx]);

    return(wld->MakeNumber(n));
}

DefineBuiltin("/", Divide, "n*n")(TSWorld * wld, int argc, TSObject * argv[])
{
    int adx;
    TSFixnum n = AsFixnum(argv[0]);

    if (argc == 1)
        return(wld->MakeNumber(0));

    for (adx = 1; adx < argc; adx++)
        n /= AsFixnum(argv[adx]);

    return(wld->MakeNumber(n));
}

DefineBuiltin("abs", Abs, "n")(TSWorld * wld, int argc, TSObject * argv[])
{
    if (AsFixnum(argv[0]) > 0)
        return(argv[0]);
    return(wld->MakeNumber(- AsFixnum(argv[0])));
}

DefineBuiltin("remainder", Remainder, "nn")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(wld->MakeNumber(AsFixnum(argv[0]) % AsFixnum(argv[1])));
}

DefineBuiltin("number->string", NumberToString, "n*n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    TSCh s[64];
    int sl;
    TSFixnum b = 10;

//    if (argc > 2)
//        wld->ThrowError("Wrong number of arguments", this);

    if (argc > 1)
        b = AsFixnum(argv[1]);

    sl = NumberAsString(AsFixnum(argv[0]), s, b);
    return(wld->MakeString(sl, s));
}

DefineBuiltin("string->number", StringToNumber, "s")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    TSFixnum n;

    if (StringAsNumber(AsString(argv[0])->Length, AsString(argv[0])->String,
            &n))
        return(wld->MakeNumber(n));
    return(wld->False);
}

//
// Booleans
//


DefineBuiltin("not", Not, ".")(TSWorld * wld, int argc, TSObject * argv[])
{
    if (argv[0] == wld->False)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("boolean?", BooleanP, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (BooleanP(argv[0]))
        return(wld->True);
    return(wld->False);
}

//
// Pairs and Lists
//

DefineBuiltin("pair?", PairP, ".")(TSWorld * wld, int argc, TSObject * argv[])
{
    if (PairP(argv[0]))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("cons", Cons, "..")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(wld->MakePair(argv[0], argv[1]));
}

DefineBuiltin("car", Car, "p")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(First(argv[0]));
}

DefineBuiltin("cdr", Cdr, "p")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(Rest(argv[0]));
}

DefineBuiltin("set-car!", SetCar, "p.")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    First(argv[0]) = argv[1];
    return(wld->Unspecified);
}

DefineBuiltin("set-cdr!", SetCdr, "p.")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    Rest(argv[0]) = argv[1];
    return(wld->Unspecified);
}

DefineBuiltin("null?", NullP, ".")(TSWorld * wld, int argc, TSObject * argv[])
{
    if (argv[0] == wld->EmptyList)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("list?", ListP, ".")(TSWorld * wld, int argc, TSObject * argv[])
{
    TSObject * fast = argv[0];
    TSObject * slow = argv[0];

    for (;;)
    {
        if (fast == wld->EmptyList)
            break;
        if (PairP(fast) == 0)
            return(wld->False);
        fast = Rest(fast);

        if (fast == wld->EmptyList)
            break;
        if (PairP(fast) == 0)
            return(wld->False);
        fast = Rest(fast);

        slow = Rest(slow);
        if (fast == slow)
            return(wld->False);
    }

    return(wld->True);
}

DefineBuiltin("list", List, "*.")(TSWorld * wld, int argc, TSObject * argv[])
{
    TSObject * obj = wld->EmptyList;
    int adx;

    for (adx = argc - 1; adx >= 0; adx--)
        obj = wld->MakePair(argv[adx], obj);

    return(obj);
}

DefineBuiltin("length", Length, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    TSObject * obj = argv[0];
    TSFixnum len = 0;

    while (obj != wld->EmptyList)
    {
        wld->MustBe(PairType, obj);
        obj = Rest(obj);
        len += 1;
    }

    return(wld->MakeNumber(len));
}

static TSObject * ListDup(TSWorld * wld, TSObject * list, TSObject * end)
{
    TSObject * ret = end;
    TSObject * obj;

    if (list == wld->EmptyList)
        return(end);

    wld->MustBe(PairType, list);
    ret = wld->MakePair(First(list), end);
    obj = ret;

    for(;;)
    {
        list = Rest(list);
        if (list == wld->EmptyList)
            break;
        wld->MustBe(PairType, list);

        Rest(obj) = wld->MakePair(First(list), end);
        obj = Rest(obj);
    }

    return(ret);
}

DefineBuiltin("append", Append, ".*.")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    int adx;
    TSObject * obj;

    if (argc == 1)
        return(argv[0]);

    obj = argv[argc - 1];
    for (adx = argc - 2; adx >= 0; adx--)
        obj = ListDup(wld, argv[adx], obj);

    return(obj);
}

static TSObject * Reverse(TSWorld * wld, TSObject * from)
{
    TSObject * to = wld->EmptyList;

    while (from != wld->EmptyList)
    {
        wld->MustBe(PairType, from);
        to = wld->MakePair(First(from), to);
        from = Rest(from);
    }

    return(to);
}

DefineBuiltin("reverse", Reverse, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(Reverse(wld, argv[0]));
}

DefineBuiltin("list-tail", ListTail, ".n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    TSObject * list = argv[0];
    TSFixnum n = AsFixnum(argv[1]);

    while (n > 0)
    {
        wld->MustBe(PairType, list);
        list = Rest(list);
        n -= 1;
    }

    return(list);
}

DefineBuiltin("list-ref", ListRef, ".n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    TSObject * list = argv[0];
    TSFixnum n = AsFixnum(argv[1]);

    while (n > 0)
    {
        wld->MustBe(PairType, list);
        list = Rest(list);
        n -= 1;
    }

    wld->MustBe(PairType, list);
    return(First(list));
}

typedef int (*EqFn)(TSObject * obj1, TSObject * obj2);

static TSObject * Member(TSWorld * wld, TSObject * obj, TSObject * list,
    EqFn fn)
{
    for (;;)
    {
        if (list == wld->EmptyList)
            break;

        wld->MustBe(PairType, list);
        if (fn(obj, First(list)))
            return(list);

        list = Rest(list);
    }

    return(wld->False);
}

DefineBuiltin("memq", Memq, "..")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(Member(wld, argv[0], argv[1], EqP));
}

DefineBuiltin("memv", Memv, "..")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(Member(wld, argv[0], argv[1], EqvP));
}

DefineBuiltin("member", Member, "..")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(Member(wld, argv[0], argv[1], EqualP));
}

static TSObject * Assoc(TSWorld * wld, TSObject * obj, TSObject * list,
    EqFn fn)
{
    for (;;)
    {
        if (list == wld->EmptyList)
            break;

        wld->MustBe(PairType, list);
        wld->MustBe(PairType, First(list));
        if (fn(obj, First(First(list))))
            return(First(list));

        list = Rest(list);
    }

    return(wld->False);
}

DefineBuiltin("assq", Assq, "..")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(Assoc(wld, argv[0], argv[1], EqP));
}

DefineBuiltin("assv", Assv, "..")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(Assoc(wld, argv[0], argv[1], EqvP));
}

DefineBuiltin("assoc", Assoc, "..")(TSWorld * wld, int argc, TSObject * argv[])
{
    return(Assoc(wld, argv[0], argv[1], EqualP));
}

//
// Symbols
//

DefineBuiltin("symbol?", SymbolP, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (SymbolP(argv[0]))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("symbol->string", SymbolToString, "S")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(AsSymbol(argv[0])->Name);
}

DefineBuiltin("string->symbol", StringToSymbol, "s")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(wld->InternSymbol(AsString(argv[0])->Length,
            AsString(argv[0])->String));
}

//
// Characters
//

DefineBuiltin("char?", CharP, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (CharacterP(argv[0]))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char=?", CharEqualP, "cc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ChCompare(AsCharacter(argv[0]), AsCharacter(argv[1])) == 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char<?", CharLessP, "cc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ChCompare(AsCharacter(argv[0]), AsCharacter(argv[1])) < 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char>?", CharGreaterP, "cc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ChCompare(AsCharacter(argv[0]), AsCharacter(argv[1])) > 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char<=?", CharLessEqualP, "cc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ChCompare(AsCharacter(argv[0]), AsCharacter(argv[1])) <= 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char>=?", CharGreaterEqualP, "cc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ChCompare(AsCharacter(argv[0]), AsCharacter(argv[1])) >= 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char-ci=?", CharCiEqualP, "cc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ChCiCompare(AsCharacter(argv[0]), AsCharacter(argv[1])) == 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char-ci<?", CharCiLessP, "cc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ChCiCompare(AsCharacter(argv[0]), AsCharacter(argv[1])) < 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char-ci>?", CharCiGreaterP, "cc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ChCiCompare(AsCharacter(argv[0]), AsCharacter(argv[1])) > 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char-ci<=?", CharCiLessEqualP, "cc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ChCiCompare(AsCharacter(argv[0]), AsCharacter(argv[1])) <= 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char-ci>=?", CharCiGreaterEqualP, "cc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ChCiCompare(AsCharacter(argv[0]), AsCharacter(argv[1])) >= 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char-alphabetic?", CharAlphabeticP, "c")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    if (ChAlphabeticP(AsCharacter(argv[0])))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char-numeric?", CharNumericP, "c")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    if (ChNumericP(AsCharacter(argv[0])))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char-whitespace?", CharWhitespaceP, "c")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    if (ChWhitespaceP(AsCharacter(argv[0])))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("char->integer", CharToInteger, "c")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(wld->MakeNumber((TSFixnum) AsCharacter(argv[0])));
}

DefineBuiltin("integer->char", IntegerToChar, "n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(wld->MakeCharacter((TSCh) AsFixnum(argv[0])));
}

DefineBuiltin("char-upcase", CharUpCase, "c")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(wld->MakeCharacter(ChUpCase(AsCharacter(argv[0]))));
}

DefineBuiltin("char-downcase", CharDownCase, "c")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(wld->MakeCharacter(ChDownCase(AsCharacter(argv[0]))));
}

//
// Strings
//

DefineBuiltin("string?", StringP, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (StringP(argv[0]))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("make-string", MakeString, "n*c")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    TSCh ch = ' ';
    int sl, sdx;
    TSObject * obj;

//    if (argc > 2)
//        wld->ThrowError("Wrong number of arguments", this);

    if (argc > 1)
        ch = AsCharacter(argv[1]);

    sl = AsFixnum(argv[0]);
    if (sl < 0)
        wld->ThrowError("string length must be positive", argv[0]);

    obj = wld->MakeString(sl, 0);
    for (sdx = 0; sdx < sl; sdx++)
        AsString(obj)->String[sdx] = ch;

    return(obj);
}

DefineBuiltin("string", String, "*c")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    int adx;
    TSObject * obj;

    obj = wld->MakeString(argc, 0);
    for (adx = 0; adx < argc; adx++)
        AsString(obj)->String[adx] = AsCharacter(argv[adx]);

    return(obj);
}

DefineBuiltin("string-length", StringLength, "s")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(wld->MakeNumber(AsString(argv[0])->Length));
}

DefineBuiltin("string-ref", StringRef, "sn")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (AsFixnum(argv[1]) >= AsString(argv[0])->Length
            || AsFixnum(argv[1]) < 0)
        wld->ThrowError("String index out of range", argv[1]);

    return(wld->MakeCharacter(AsString(argv[0])->String[ AsFixnum(argv[1])]));
}

DefineBuiltin("string-set!", StringSet, "snc")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (AsFixnum(argv[1]) >= AsString(argv[0])->Length
            || AsFixnum(argv[1]) < 0)
        wld->ThrowError("String index out of range", argv[1]);

    AsString(argv[0])->String[ AsFixnum(argv[1])] = AsCharacter(argv[2]);
    return(wld->Unspecified);
}

static int StringCompare(TSString * str1, TSString * str2)
{
    int sdx;

    for (sdx = 0; sdx < str1->Length && sdx < str2->Length; sdx++)
        if (str1->String[sdx] != str2->String[sdx])
            return(str1->String[sdx] - str2->String[sdx]);

    return(str1->Length - str2->Length);
}

static int StringCiCompare(TSString * str1, TSString * str2)
{
    int sdx;

    for (sdx = 0; sdx < str1->Length && sdx < str2->Length; sdx++)
        if (ChCiCompare(str1->String[sdx], str2->String[sdx]) != 0)
            return(ChCiCompare(str1->String[sdx], str2->String[sdx]));

    return(str1->Length - str2->Length);
}

DefineBuiltin("string=?", StringEqualP, "ss")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (StringCompare(AsString(argv[0]), AsString(argv[1])) == 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("string<?", StringLessP, "ss")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (StringCompare(AsString(argv[0]), AsString(argv[1])) < 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("string>?", StringGreaterP, "ss")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (StringCompare(AsString(argv[0]), AsString(argv[1])) > 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("string<=?", StringLessEqualP, "ss")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (StringCompare(AsString(argv[0]), AsString(argv[1])) <= 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("string>=?", StringGreaterEqualP, "ss")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (StringCompare(AsString(argv[0]), AsString(argv[1])) >= 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("string-ci=?", StringCiEqualP, "ss")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (StringCiCompare(AsString(argv[0]), AsString(argv[1])) == 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("string-ci<?", StringCiLessP, "ss")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (StringCiCompare(AsString(argv[0]), AsString(argv[1])) < 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("string-ci>?", StringCiGreaterP, "ss")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (StringCiCompare(AsString(argv[0]), AsString(argv[1])) > 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("string-ci<=?", StringCiLessEqualP, "ss")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    if (StringCiCompare(AsString(argv[0]), AsString(argv[1])) <= 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("string-ci>=?", StringCiGreaterEqualP, "ss")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    if (StringCiCompare(AsString(argv[0]), AsString(argv[1])) >= 0)
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("substring", Substring, "snn")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    int start = AsFixnum(argv[1]);
    int end = AsFixnum(argv[2]);

    if (start < 0 || start > end)
        wld->ThrowError("bad string index", argv[1]);
    if (end > AsString(argv[0])->Length)
        wld->ThrowError("bad string index", argv[2]);

    return(wld->MakeString(end - start, AsString(argv[0])->String + start));
}

DefineBuiltin("string-append", StringAppend, "*s")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    int sl;
    TSObject * obj;
    int adx;
    int sdx;

    for (sl = 0, adx = 0; adx < argc; adx++)
        sl += AsString(argv[adx])->Length;

    obj = wld->MakeString(sl, 0);

    for (sl = 0, adx = 0; adx < argc; adx++)
    {
        for (sdx = 0; sdx < AsString(argv[adx])->Length; sdx++)
            AsString(obj)->String[sl + sdx] = AsString(argv[adx])->String[sdx];
        sl += AsString(argv[adx])->Length;
    }

    return(obj);
}

DefineBuiltin("string->list", StringToList, "s")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    int sdx;
    TSObject * obj = wld->EmptyList;

    for (sdx = AsString(argv[0])->Length - 1; sdx >= 0; sdx--)
        obj = wld->MakePair(wld->MakeCharacter(AsString(argv[0])->String[sdx]),
                obj);

    return(obj);
}

DefineBuiltin("list->string", ListToString, ".")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    TSObject * list = argv[0];
    int sl = ListLength(wld, list);
    TSObject * obj = wld->MakeString(sl, 0);
    int sdx;

    for (sdx = 0; sdx < sl; sdx++)
    {
        wld->MustBe(CharacterType, First(list));
        AsString(obj)->String[sdx] = AsCharacter(First(list));
        list = Rest(list);
    }

    return(obj);
}

DefineBuiltin("string-copy", StringCopy, "s")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    return(wld->MakeString(AsString(argv[0])->Length,
            AsString(argv[0])->String));
}

DefineBuiltin("string-fill!", StringFill, "sc")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    int sdx;
    TSCh ch = AsCharacter(argv[1]);

    for (sdx = 0; sdx < AsString(argv[0])->Length; sdx++)
        AsString(argv[0])->String[sdx] = ch;

    return(wld->Unspecified);
}

//
// Vectors
//

DefineBuiltin("vector?", VectorP, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (VectorP(argv[0]))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("make-vector", MakeVector, "n*.")(TSWorld * wld, int argc,
    TSObject * argv[])
{
//    if (argc > 2)
//        wld->ThrowError("Wrong number of arguments", this);

    if (argc > 1)
        return(wld->MakeVector(AsFixnum(argv[0]), argv[1]));
    return(wld->MakeVector(AsFixnum(argv[0]), wld->Unspecified));
}

DefineBuiltin("vector", Vector, "*.")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    TSObject * obj = wld->MakeVector(argc, 0);
    int adx;

    for (adx = 0; adx < argc; adx++)
        AsVector(obj)->Vector[adx] = argv[adx];

    return(obj);
}

DefineBuiltin("vector-length", VectorLength, "v")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(wld->MakeNumber(AsVector(argv[0])->Length));
}

DefineBuiltin("vector-ref", VectorRef, "vn")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (AsFixnum(argv[1]) >= AsVector(argv[0])->Length
            || AsFixnum(argv[1]) < 0)
        wld->ThrowError("Vector index out of range", argv[1]);

    return(AsVector(argv[0])->Vector[AsFixnum(argv[1])]);
}

DefineBuiltin("vector-set!", VectorSet, "vn.")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (AsFixnum(argv[1]) >= AsVector(argv[0])->Length
            || AsFixnum(argv[1]) < 0)
        wld->ThrowError("Vector index out of range", argv[1]);

    AsVector(argv[0])->Vector[AsFixnum(argv[1])] = argv[2];
    return(wld->Unspecified);
}

DefineBuiltin("vector->list", VectorToList, "v")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    TSObject * obj = wld->EmptyList;
    int vdx;

    for (vdx = AsVector(argv[0])->Length - 1; vdx >= 0; vdx--)
        obj = wld->MakePair(AsVector(argv[0])->Vector[vdx], obj);
    return(obj);
}

DefineBuiltin("list->vector", ListToVector, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    TSObject * list = argv[0];
    int vl = ListLength(wld, list);
    TSObject * obj = wld->MakeVector(vl, 0);
    int vdx;

    for (vdx = 0; vdx < vl; vdx++)
    {
        AsVector(obj)->Vector[vdx] = First(list);
        list = Rest(list);
    }

    return(obj);
}

DefineBuiltin("vector-fill!", VectorFill, "v.")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    int vdx;

    for (vdx = 0; vdx < AsVector(argv[0])->Length; vdx++)
        AsVector(argv[0])->Vector[vdx] = argv[1];

    return(wld->Unspecified);
}

//
// Control Features
//

DefineBuiltin("procedure?", ProcedureP, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    if (ProcedureP(argv[0]) || BuiltinP(argv[0]))
        return(wld->True);
    return(wld->False);
}

DefineBuiltin("apply", Apply, "..*.")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    int ac = 0;
    int adx;
    TSObject * av[128];
    TSObject * args;

    if (ProcedureP(argv[0]) == 0 && BuiltinP(argv[0]) == 0)
        wld->ThrowError("Not a callable", argv[0]);

    for (adx = 1; adx < argc - 1; adx++)
        av[adx - 1] = argv[adx];

    args = argv[argc - 1];
    ac = argc - 2;
    while (args != wld->EmptyList)
    {
        wld->MustBe(PairType, args);
        av[ac] = First(args);
        args = Rest(args);

        ac += 1;
        if (ac == sizeof(av) / sizeof(TSObject *))
            wld->ThrowError("Too many arguments", argv[0]);
    }

    return(argv[0]->Call(wld, ac, av));
}

static TSObject * ForEachMap(TSWorld * wld, int argc, TSObject * argv[])
{
    TSObject * lists[128];
    TSObject * av[128];
    int adx;
    TSObject * obj = wld->EmptyList;
    int ll;

    lists[0] = argv[1];
    ll = ListLength(wld, lists[0]);
    for (adx = 2; adx < argc; adx++)
    {
        lists[adx - 1] = argv[adx];
        if (ListLength(wld, lists[adx - 1]) != ll)
            wld->ThrowError("List lengths differ", lists[adx - 1]);
    }

    while (lists[0] != wld->EmptyList)
    {
        for (adx = 0; adx < argc - 1; adx++)
        {
            av[adx] = First(lists[adx]);
            lists[adx] = Rest(lists[adx]);
        }

// GC: need to protect obj

        obj = wld->MakePair(Evaluate(wld, argv[0]->Call(wld, argc - 1, av)),
                obj);
    }

    return(obj);
}

DefineBuiltin("map", Map, "..*.")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    return(Reverse(wld, ForEachMap(wld, argc, argv)));
}

DefineBuiltin("for-each", ForEach, "..*.")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    ForEachMap(wld, argc, argv);
    return(wld->Unspecified);
}

//
// Input and Output
//



DefineBuiltin("write", Write, ".")(TSWorld * wld, int argc, TSObject * argv[])
{
    wld->Write(wld->DefaultOutput, argv[0]);
    return(wld->Unspecified);
}

DefineBuiltin("display", Display, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    wld->Display(wld->DefaultOutput, argv[0]);
    return(wld->Unspecified);
}

DefineBuiltin("newline", Newline, "")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    wld->DefaultOutput->PutString("\n");
    return(wld->Unspecified);
}

TSSyntaxInfo::TSSyntaxInfo(char * name, TSSyntaxFn fn)
{
    Name = name;
    Syntax = fn;
    Next = SyntaxInfoList;
    SyntaxInfoList = this;
}

TSObject * TSSyntax::Syntax(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    return(SyntaxFn(wld, expr, obj));
}

TSSyntax::TSSyntax(TSWorld * wld, char * name, TSSyntaxFn fn)
    : TSObject(wld, SyntaxType)
{
    Name = name;
    SyntaxFn = fn;
}

DefineSyntax("quote", Quote)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    // (quote <obj>)

    wld->MustBe(PairType, expr);
    if (Rest(expr) != wld->EmptyList)
        wld->ThrowError("Bad quote syntax", obj);
    return(First(expr));
}

DefineSyntax("define", Define)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    wld->MustBe(PairType, expr);
    wld->MustBe(PairType, Rest(expr));

    if (wld->Stack->AsEvaluateContext() == 0)
        wld->ThrowError("Define not allowed here", obj);

    if (SymbolP(First(expr)))
    {
        // (define <symbol> <expression>)

        if (Rest(Rest(expr)) != wld->EmptyList)
            wld->ThrowError("Bad define syntax", obj);

        AsSymbol(First(expr))->Value = Evaluate(wld, First(Rest(expr)));

        return(wld->Unspecified);
    }

    // (define (<symbol> ...) <expression> ...)

    wld->MustBe(PairType, First(expr));
    wld->MustBe(SymbolType, First(First(expr)));

    AsSymbol(First(First(expr)))->Value = wld->MakeProcedure(
            First(First(expr)), Rest(First(expr)), Rest(expr));
    return(wld->Unspecified);
}

DefineSyntax("lambda", Lambda)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    // (lambda <formals> <body>)

    wld->MustBe(PairType, expr);
    return(wld->MakeProcedure(wld->Unspecified, First(expr), Rest(expr)));
}

DefineSyntax("if", If)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    // (if <test> <consequent>)
    // (if <test> <consequent> <alternate>)

    wld->MustBe(PairType, expr);
    wld->MustBe(PairType, Rest(expr));

    if (Evaluate(wld, First(expr)) == wld->False)
    {
        if (Rest(Rest(expr)) == wld->EmptyList)
            return(wld->Unspecified);

        wld->MustBe(PairType, Rest(Rest(expr)));
        return(EvaluateBlock(wld, Rest(Rest(expr))));
    }

    return(First(Rest(expr))->Evaluate(wld));
}

DefineSyntax("set!", Set)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    // (set! <variable> <expression>)

    wld->MustBe(PairType, expr);
    wld->MustBe(PairType, Rest(expr));
    wld->MustBe(SymbolType, First(expr));
    if (Rest(Rest(expr)) != wld->EmptyList)
        wld->ThrowError("Bad set! syntax", obj);
    AsSymbol(First(expr))->SetValue(wld, Evaluate(wld, First(Rest(expr))));

    return(wld->Unspecified);
}

DefineSyntax("begin", Begin)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    // (begin <expression> ...)

    wld->MustBe(PairType, expr);

    return(EvaluateBlock(wld, expr));
}

DefineSyntax("and", And)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    // (and <test> ...)

    TSObject * ret = wld->True;

    while (expr != wld->EmptyList)
    {
        wld->MustBe(PairType, expr);
        ret = Evaluate(wld, First(expr));
        if (ret == wld->False)
            return(ret);
        expr = Rest(expr);
    }

    return(ret);
}

DefineSyntax("or", Or)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    // (or <test> ...)

    TSObject * ret = wld->False;

    while (expr != wld->EmptyList)
    {
        wld->MustBe(PairType, expr);
        ret = Evaluate(wld, First(expr));
        if (ret != wld->False)
            return(ret);
        expr = Rest(expr);
    }

    return(ret);
}

DefineSyntax("cond", Cond)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    // (cond <clause> ...)
    // <clause> := (<test> <expression> ...)
    // <clause> := (<test>)
    // <clause> := (<test> => <expression>)
    // last <clause> := (else <expression> ...)

    TSObject * clause;

    while (expr != wld->EmptyList)
    {
        wld->MustBe(PairType, expr);
        clause = First(expr);
        wld->MustBe(PairType, clause);
        if (First(clause) == wld->SymbolElse)
        {
            if (Rest(expr) != wld->EmptyList)
                wld->ThrowError("Else must be last clause in cond", obj);

            return(EvaluateBlock(wld, Rest(clause)));
        }

        if (First(Rest(clause)) == wld->SymbolEqualArrow)
        {
            wld->MustBe(PairType, Rest(clause));
            wld->MustBe(PairType, Rest(Rest(clause)));
            if (Rest(Rest(Rest(clause))) != wld->EmptyList)
                wld->ThrowError("Bad cond syntax", obj);

            TSObject * val = Evaluate(wld, First(clause));

// GC: need to protect val

            if (val != wld->False)
                return(Evaluate(wld, First(Rest(Rest(clause))))->Call(wld,
                        1, &val));
        }
        else
        {
            TSObject * val = Evaluate(wld, First(clause));

            if (val != wld->False)
            {
                if (Rest(clause) == wld->EmptyList)
                    return(val);
                return(EvaluateBlock(wld, Rest(clause)));
            }
        }

        expr = Rest(expr);
    }

    return(wld->Unspecified);
}

DefineSyntax("case", Case)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    // (case <key> <clause> ...)
    // <clause> := ((<datum> ...) <expression> ...)
    // last <clause> := (else <expression> ...)

    TSObject * clause;
    TSObject * key;

    wld->MustBe(PairType, expr);

// GC: need to protect key

    key = Evaluate(wld, First(expr));
    expr = Rest(expr);
    while (expr != wld->EmptyList)
    {
        wld->MustBe(PairType, expr);
        clause = First(expr);
        wld->MustBe(PairType, clause);
        wld->MustBe(PairType, Rest(clause));
        if (First(clause) == wld->SymbolElse)
        {
            if (Rest(expr) != wld->EmptyList)
                wld->ThrowError("Else must be last clause in cond", obj);

            return(EvaluateBlock(wld, Rest(clause)));
        }

        if (Member(wld, key, First(clause), EqvP) != wld->False)
            return(EvaluateBlock(wld, Rest(clause)));

        expr = Rest(expr);
    }

    return(wld->Unspecified);
}

// (Internal) DefineContext

class TSDefineContext : public TSContext
{
public:

    TSObject * Defines;
    TSObject ** Values;
    TSObject * Body;

    virtual TSObject * Evaluate(TSWorld * wld);
    virtual TSObject * GetValue(TSWorld * wld, TSSymbol * sym);
    virtual int SetValue(TSWorld * wld, TSSymbol * sym, TSObject * val);
    virtual TSLineNumberMap * GetLineNumberMap();
    virtual TSProcedure * GetProcedure();

    TSDefineContext(TSWorld * wld, TSObject * defs, int argc, TSObject * body);
    ~TSDefineContext();
};

TSObject * TSDefineContext::Evaluate(TSWorld * wld)
{
    int idx;
    TSObject * obj;
    TSObject * def;

    Stack = wld->Stack;
    wld->Stack = this;

    for (idx = 0, obj = Defines; obj != wld->EmptyList; obj = Rest(obj),
            idx++)
    {
        def = First(obj);
        if (PairP(def) == 0 || First(def) != wld->SymbolDefine)
            break;

        def = Rest(def);
        if (SymbolP(First(def)))
            Values[idx] = ::Evaluate(wld, First(Rest(def)));
        else
            Values[idx] = wld->MakeProcedure(First(First(def)),
                    Rest(First(def)), Rest(def));
    }

    obj = EvaluateBlock(wld, Body);
    wld->Stack = Stack;

    return(obj);
}

TSObject * TSDefineContext::GetValue(TSWorld * wld, TSSymbol * sym)
{
    int idx;
    TSObject * obj;
    TSObject * def;

    for (idx = 0, obj = Defines; obj != wld->EmptyList; obj = Rest(obj),
            idx++)
    {
        def = First(obj);
        if (PairP(def) == 0 || First(def) != wld->SymbolDefine)
            break;

        def = Rest(def);
        if (First(def) == sym || (PairP(First(def))
                && First(First(def)) == sym))
            return(Values[idx]);
    }

    return(Stack->GetValue(wld, sym));
}

int TSDefineContext::SetValue(TSWorld * wld, TSSymbol * sym, TSObject * val)
{
    int idx;
    TSObject * obj;
    TSObject * def;

    for (idx = 0, obj = Defines; obj != wld->EmptyList; obj = Rest(obj),
            idx++)
    {
        def = First(obj);
        if (PairP(def) == 0 || First(def) != wld->SymbolDefine)
            break;

        def = Rest(def);
        if (First(def) == sym || (PairP(First(def))
                && First(First(def)) == sym))
        {
            Values[idx] = val;
            return(1);
        }
    }

    return(Stack->SetValue(wld, sym, val));
}

TSLineNumberMap * TSDefineContext::GetLineNumberMap()
{
    return(Stack->GetLineNumberMap());
}

TSProcedure * TSDefineContext::GetProcedure()
{
    return(Stack->GetProcedure());
}

TSDefineContext::TSDefineContext(TSWorld * wld, TSObject * defs, int argc,
    TSObject * body) : TSContext(wld, body)
{
    int adx;

    Defines = defs;
    Body = body;
    Values = new TSObject * [argc];
    for (adx = 0; adx < argc; adx++)
        Values[adx] = wld->Unspecified;
}

TSDefineContext::~TSDefineContext()
{
    if (Values != 0)
        delete Values;
}

static TSDefineContext * InternalDefinesDo(TSWorld * wld, TSObject * body)
{
    // (<definition> ... <expression> ...)
    // <definition> := (define <symbol> <expression>)
    // <definition> := (define (<symbol> ...) <expression> ...)

    int argc = 0;
    TSObject * def;
    TSObject * defs = body;

    while (body != wld->EmptyList)
    {
        wld->MustBe(PairType, body);
        def = First(body);
        if (PairP(def) == 0 || First(def) != wld->SymbolDefine)
            break;

        def = Rest(def);
        wld->MustBe(PairType, def);
        wld->MustBe(PairType, Rest(def));

        if (SymbolP(First(def)))
        {
            // (define <symbol> <expression>)

            if (Rest(Rest(def)) != wld->EmptyList)
                wld->ThrowError("Bad define syntax", First(body));
        }
        else
        {
            // (define (<symbol> ...) <expression> ...)

            wld->MustBe(PairType, First(def));
            wld->MustBe(SymbolType, First(First(def)));
        }

        argc += 1;
        body = Rest(body);
    }

    return(new TSDefineContext(wld, defs, argc, body));
}

static TSDefineContext * InternalDefines(TSWorld * wld, TSObject * body)
{
    if (PairP(body) && PairP(First(body))
            && First(First(body)) == wld->SymbolDefine)
        return(InternalDefinesDo(wld, body));

    return(0);
}

class TSLetContext : public TSContext
{
public:

    TSObject * Variables;
    TSObject ** Values;
    TSObject * Body;
    int LetRecFlag;

    virtual TSObject * Evaluate(TSWorld * wld);
    virtual TSObject * GetValue(TSWorld * wld, TSSymbol * sym);
    virtual int SetValue(TSWorld * wld, TSSymbol * sym, TSObject * val);
    virtual TSLineNumberMap * GetLineNumberMap();
    virtual TSProcedure * GetProcedure();

    TSLetContext(TSWorld * wld, TSObject * vars, int argc,
            TSObject * argv[], TSObject * body, int lrf);
    ~TSLetContext();
};

TSObject * TSLetContext::Evaluate(TSWorld * wld)
{
    TSObject * obj;
    TSDefineContext * ctx;

    Stack = wld->Stack;
    wld->Stack = this;
    if (LetRecFlag)
    {
        for (obj = Variables; obj != wld->EmptyList; obj = Rest(obj))
            AsSymbol(First(First(obj)))->SetValue(wld,
                    ::Evaluate(wld, First(Rest(First(obj)))));
    }

    ctx = InternalDefines(wld, Body);
    if (ctx == 0)
        obj = EvaluateBlock(wld, Body);
    else
        obj = ctx->Evaluate(wld);

    wld->Stack = Stack;

    return(obj);
}

TSObject * TSLetContext::GetValue(TSWorld * wld, TSSymbol * sym)
{
    int idx;
    TSObject * obj;

    for (idx = 0, obj = Variables; obj != wld->EmptyList; obj = Rest(obj),
            idx++)
        if (First(First(obj)) == sym)
            return(Values[idx]);

    return(Stack->GetValue(wld, sym));
}

int TSLetContext::SetValue(TSWorld * wld, TSSymbol * sym, TSObject * val)
{
    int idx;
    TSObject * obj;

    for (idx = 0, obj = Variables; obj != wld->EmptyList; obj = Rest(obj),
            idx++)
        if (First(First(obj)) == sym)
        {
            Values[idx] = val;
            return(1);
        }

    return(Stack->SetValue(wld, sym, val));
}

TSLineNumberMap * TSLetContext::GetLineNumberMap()
{
    return(Stack->GetLineNumberMap());
}

TSProcedure * TSLetContext::GetProcedure()
{
    return(Stack->GetProcedure());
}

TSLetContext::TSLetContext(TSWorld * wld, TSObject * vars, int argc,
    TSObject * argv[], TSObject * body, int lrf) : TSContext(wld, body)
{
    int adx;

    Variables = vars;
    Body = body;
    Values = new TSObject * [argc];
    for (adx = 0; adx < argc; adx++)
        Values[adx] = argv[adx];
    LetRecFlag = lrf;
}

TSLetContext::~TSLetContext()
{
    if (Values != 0)
        delete Values;
}

static TSObject * LetAndLetRecSyntax(TSWorld * wld, TSObject * expr,
    TSObject * obj, int lrf)
{
    // (let ((<variable> <init>) ...) <expression> ...)
    // (letrec ((<variable> <init>) ...) <expression> ...)

    TSObject * bindings;
    int argc = 0;
    TSObject * argv[128];
    TSContext * ctx;

    wld->MustBe(PairType, expr);
    wld->MustBe(PairType, Rest(expr));
    wld->MustBe(PairType, First(expr));

    for (bindings = First(expr); bindings != wld->EmptyList;
            bindings = Rest(bindings))
    {
        wld->MustBe(PairType, bindings);
        wld->MustBe(PairType, First(bindings));
        wld->MustBe(SymbolType, First(First(bindings)));
        wld->MustBe(PairType, Rest(First(bindings)));
        if (Rest(Rest(First(bindings))) != wld->EmptyList)
            wld->ThrowError("Bad let syntax", obj);

        if (argc == sizeof(argv) / sizeof(TSObject *))
            wld->ThrowError("Too many let variables", obj);

// GC: need to protect argv

        if (lrf)
            argv[argc] = wld->Unspecified;
        else
            argv[argc] = Evaluate(wld, First(Rest(First(bindings))));

        argc += 1;
    }

    ctx = new TSLetContext(wld, First(expr), argc, argv, Rest(expr), lrf);
    return(ctx->Evaluate(wld));
}

DefineSyntax("let", Let)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    return(LetAndLetRecSyntax(wld, expr, obj, 0));
}

DefineSyntax("letrec", LetRec)(TSWorld * wld, TSObject * expr, TSObject * obj)
{
    return(LetAndLetRecSyntax(wld, expr, obj, 1));
}

TSContext::TSContext(TSWorld * wld, TSObject * expr)
    : TSObject(wld, ContextType)
{
    Stack = 0;
    Expression = expr;
}

TSContext::~TSContext()
{
    // Nothing.
}

TSEvaluateContext * TSContext::AsEvaluateContext()
{
    return(0);
}

class TSCallContext : public TSContext
{
public:

    TSProcedure * Procedure;
    TSObject ** Values;

    virtual TSObject * Evaluate(TSWorld * wld);
    virtual TSObject * GetValue(TSWorld * wld, TSSymbol * sym);
    virtual int SetValue(TSWorld * wld, TSSymbol * sym, TSObject * val);
    virtual TSLineNumberMap * GetLineNumberMap();
    virtual TSProcedure * GetProcedure();

    TSCallContext(TSWorld * wld, TSProcedure * proc, int argc,
            TSObject * argv[]);
    ~TSCallContext();
};

TSObject * TSCallContext::Evaluate(TSWorld * wld)
{
    TSObject * obj;
    TSDefineContext * ctx;

    Stack = wld->Stack;
    wld->Stack = this;

    ctx = InternalDefines(wld, Procedure->Body);
    if (ctx == 0)
        obj = EvaluateBlock(wld, Procedure->Body);
    else
        obj = ctx->Evaluate(wld);

    wld->Stack = Stack;
    Stack = 0;

    return(obj);
}

TSObject * TSCallContext::GetValue(TSWorld * wld, TSSymbol * sym)
{
    int idx;
    TSObject * obj;

    for (idx = 0, obj = Procedure->Formals; obj != wld->EmptyList;
            obj = Rest(obj), idx++)
    {
        if (SymbolP(obj))
        {
            if (obj == sym)
                return(Values[idx]);

            break;
        }

        if (First(obj) == sym)
            return(Values[idx]);
    }

    if (Procedure->Enclosing == 0)
        return(Undefined);

    return(Procedure->Enclosing->GetValue(wld, sym));
}

int TSCallContext::SetValue(TSWorld * wld, TSSymbol * sym, TSObject * val)
{
    int idx;
    TSObject * obj;

    for (idx = 0, obj = Procedure->Formals; obj != wld->EmptyList;
            obj = Rest(obj), idx++)
    {
        if (SymbolP(obj))
        {
            if (obj == sym)
            {
                Values[idx] = val;
                return(1);
            }

            break;
        }

        if (First(obj) == sym)
        {
            Values[idx] = val;
            return(1);
        }
    }

    if (Procedure->Enclosing == 0)
        return(0);

    return(Procedure->Enclosing->SetValue(wld, sym, val));
}

TSLineNumberMap * TSCallContext::GetLineNumberMap()
{
    return(Procedure->LineNumberMap);
}

TSProcedure * TSCallContext::GetProcedure()
{
    return(Procedure);
}

TSCallContext::TSCallContext(TSWorld * wld, TSProcedure * proc, int argc,
    TSObject * argv[]) : TSContext(wld, proc->Body)
{
    int adx;

    Procedure = proc;
    Values = 0;

    if (Procedure->RestFlag)
    {
        if (argc < Procedure->ArgCount - 1)
            wld->ThrowError("Wrong number of arguments", proc);

        Values = new TSObject * [Procedure->ArgCount];
        for (adx = 0; adx < Procedure->ArgCount - 1; adx++)
            Values[adx] = argv[adx];

        Values[Procedure->ArgCount - 1] = wld->EmptyList;
        for (adx = argc - 1; adx >= Procedure->ArgCount - 1; adx--)
            Values[Procedure->ArgCount - 1] = wld->MakePair(argv[adx],
                    Values[Procedure->ArgCount - 1]);
    }
    else
    {
        if (argc != Procedure->ArgCount)
            wld->ThrowError("Wrong number of arguments", proc);

        Values = new TSObject * [Procedure->ArgCount];
        for (adx = 0; adx < argc; adx++)
            Values[adx] = argv[adx];
    }
}

TSCallContext::~TSCallContext()
{
    if (Values != 0)
        delete Values;
}

TSObject * TSProcedure::Call(TSWorld * wld, int argc, TSObject * argv[])
{
    return(new TSCallContext(wld, this, argc, argv));
}

TSProcedure::TSProcedure(TSWorld * wld, TSObject * name, TSObject * formals,
    TSObject * body, TSLineNumberMap * lnm) : TSObject(wld, ProcedureType)
{
    TSObject * obj;

    Name = name;
    Formals = formals;
    ArgCount = 0;
    RestFlag = 0;
    Body = body;
    Enclosing = wld->Stack;

    if (lnm)
        LineNumberMap = new TSLineNumberMap(lnm);
    else
        LineNumberMap = 0;

    obj = formals;
    for (;;)
    {
        if (SymbolP(obj))
        {
            ArgCount += 1;
            RestFlag = 1;
            break;
        }
        else if (obj == wld->EmptyList)
            break;
        else if (PairP(obj) == 0 || SymbolP(First(obj)) == 0)
            wld->ThrowError("Bad formal list for lambda", formals);

        ArgCount += 1;
        obj = Rest(obj);
    }
}

TSProcedure::~TSProcedure()
{
    if (LineNumberMap)
        delete LineNumberMap;
}

class TSUnspecified : public TSObject
{
private:

    friend class TSWorld;

    TSUnspecified(TSWorld * wld) : TSObject(wld, UnspecifiedType)
        { /* Nothing. */ }
};

class TSEmptyList : public TSObject
{
public:

    virtual TSObject * Evaluate(TSWorld * wld);

private:

    friend class TSWorld;

    TSEmptyList(TSWorld * wld) : TSObject(wld, EmptyListType)
        { /* Nothing. */ }
};

TSObject * TSEmptyList::Evaluate(TSWorld * wld)
{
    wld->ThrowError("The empty list must be quoted", this);
    return(this);
}

class TSEndOfFile : public TSObject
{
private:

    friend class TSWorld;

    TSEndOfFile(TSWorld * wld) : TSObject(wld, EndOfFileType)
        { /* Nothing. */ }
};

TSLineNumberMap::TSLineNumberMap(TSCh * fn)
{
    Filename = StringDup(StringLength(fn), fn);
    MapLength = 0;
    MapSize = 8;
    Map = new TSLineNumber[MapSize];
}

TSLineNumberMap::TSLineNumberMap(TSLineNumberMap * lnm)
{
    int mdx;

    Filename = StringDup(StringLength(lnm->Filename), lnm->Filename);
    MapLength = lnm->MapLength;
    MapSize = MapLength;
    Map = new TSLineNumber[MapSize];

    for (mdx = 0; mdx < MapLength; mdx++)
        Map[mdx] = lnm->Map[mdx];
}

TSLineNumberMap::~TSLineNumberMap()
{
    delete Filename;
    delete Map;
}

TSObject * TSLineNumberMap::Record(TSObject * obj, int ln)
{
    TSLineNumber * nm;
    int mdx;

    if (MapLength == MapSize)
    {
        nm = new TSLineNumber[MapSize * 2];
        for (mdx = 0; mdx < MapSize; mdx++)
            nm[mdx] = Map[mdx];
        delete Map;
        Map = nm;
        MapSize *= 2;
    }

    Map[MapLength].LineNumber = ln;
    Map[MapLength].Object = obj;
    MapLength += 1;

    return(obj);
}

TSWorld::TSWorld(TSOutputStream * os, TSApplication * app)
{
    unsigned int idx;

    DefaultOutput = os;
    Application = app;
    Objects = 0;
    Stack = 0;

    for (idx = 0; idx < sizeof(Symbols) / sizeof(TSSymbol *); idx++)
        Symbols[idx] = 0;

    False = new TSBoolean(this);
    True = new TSBoolean(this);
    Unspecified = new TSUnspecified(this);
    EmptyList = new TSEmptyList(this);
    EndOfFile = new TSEndOfFile(this);

    SymbolQuote = InternSymbol("quote");
    SymbolEqualArrow = InternSymbol("=>");
    SymbolElse = InternSymbol("else");
    SymbolDefine = InternSymbol("define");

    AddAllBuiltins(this);
    AddAllSyntax(this);
}

TSWorld::~TSWorld()
{
    TSObject * obj;

    while (Objects != 0)
    {
        obj = Objects;
        Objects = Objects->Next;
        delete obj;
    }
}

class TSEvaluateContext : public TSContext
{
public:

    TSObject * EvaluateExpression;
    TSLineNumberMap * LineNumberMap;

    virtual TSObject * Evaluate(TSWorld * wld);
    virtual TSObject * GetValue(TSWorld * wld, TSSymbol * sym);
    virtual int SetValue(TSWorld * wld, TSSymbol * sym, TSObject * val);
    virtual TSLineNumberMap * GetLineNumberMap();
    virtual TSProcedure * GetProcedure();
    virtual TSEvaluateContext * AsEvaluateContext();

    TSEvaluateContext(TSWorld * wld, TSObject * expr, TSLineNumberMap * lnm);
};

TSObject * TSEvaluateContext::Evaluate(TSWorld * wld)
{
    TSObject * obj;

    Stack = wld->Stack;
    wld->Stack = this;
    obj = ::Evaluate(wld, EvaluateExpression);
    wld->Stack = Stack;
    Stack = 0;

    return(obj);
}

TSObject * TSEvaluateContext::GetValue(TSWorld * wld, TSSymbol * sym)
{
    return(Undefined);
}

int TSEvaluateContext::SetValue(TSWorld * wld, TSSymbol * sym, TSObject * val)
{
    return(0);
}

TSLineNumberMap * TSEvaluateContext::GetLineNumberMap()
{
    return(LineNumberMap);
}

TSProcedure * TSEvaluateContext::GetProcedure()
{
    return(0);
}

TSEvaluateContext * TSEvaluateContext::AsEvaluateContext()
{
    return(this);
}

TSEvaluateContext::TSEvaluateContext(TSWorld * wld, TSObject * expr,
    TSLineNumberMap * lnm) : TSContext(wld, expr)
{
    EvaluateExpression = expr;
    LineNumberMap = lnm;
}

TSObject * TSWorld::Evaluate(TSObject * obj, TSLineNumberMap * lnm)
{
    TSContext * ctx = new TSEvaluateContext(this, obj, lnm);
    return(ctx->Evaluate(this));
}

void TSWorld::ThrowError(char * es, TSObject * obj)
{
    TSContext * ctx;
    int mdx;
    TSCh s[64];
    int sl;
    TSProcedure * prev = 0;

    DefaultOutput->PutString(es);
    DefaultOutput->PutString(": ");
    Write(DefaultOutput, obj);
    DefaultOutput->PutString("\n\n");

    for (ctx = Stack; ctx != 0; ctx = ctx->Stack)
    {
        TSProcedure * proc = ctx->GetProcedure();

        if (proc != 0)
        {
            if (proc != prev)
            {
                if (proc->Name != Unspecified)
                {
                    Write(DefaultOutput, proc->Name);
                    DefaultOutput->PutString(" ");
                }

                Write(DefaultOutput, proc);
                DefaultOutput->PutString("\n");

                prev = proc;
            }

        }
        else
        {
            DefaultOutput->PutString("eval ");
            Write(DefaultOutput, ctx->AsEvaluateContext()->EvaluateExpression);
            DefaultOutput->PutString("\n");
        }

        DefaultOutput->PutString("    ");

        TSLineNumberMap * lnm = ctx->GetLineNumberMap();
        if (lnm)
        {
            DefaultOutput->PutString(lnm->Filename,
                    StringLength(lnm->Filename));
            for (mdx = 0; mdx < lnm->MapLength; mdx++)
            {
                if (ctx->Expression == lnm->Map[mdx].Object)
                {
                    sl = NumberAsString(lnm->Map[mdx].LineNumber, s, 10);
                    DefaultOutput->PutString("(");
                    DefaultOutput->PutString(s, sl);
                    DefaultOutput->PutString(")");
                    break;
                }
            }

            DefaultOutput->PutString(": ");
        }

        Write(DefaultOutput, ctx->Expression);
        DefaultOutput->PutString("\n");
    }

    Stack = 0;
    Application->FatalError();
}

void TSWorld::MustBe(TSObjectType ot, TSObject * obj)
{
    if (ObjectType(obj) != ot)
        ThrowError("Wrong type", obj);
}

TSObject * TSWorld::Read(TSInputStream * is, TSLineNumberMap * lnm)
{
    return(ReadInternal(is, lnm, 1, 0));
}

#define SymbolChP(ch) \
    (ChAlphabeticP((ch)) || ChNumericP((ch)) \
        || (ch) == '+' || (ch) == '-' || (ch) == '.' \
        || (ch) == '*' || (ch) == '/' || (ch) == '<' || (ch) == '=' \
        || (ch) == '>' || (ch) == '!' || (ch) == '?' || (ch) == ':' \
        || (ch) == '$' || (ch) == '%' || (ch) == '_' || (ch) == '&' \
        || (ch) == '~' || (ch) == '^')

#define DotObj ((TSObject *) -1)
#define EolObj ((TSObject *) -2)

TSObject * TSWorld::ReadInternal(TSInputStream * is, TSLineNumberMap * lnm,
    int eaf, int rlf)
{
    TSCh ch;
    int eof;

    for (;;)
    {
        ch = is->GetCh(&eof);
        if (eof)
            goto Eof;

        switch (ch)
        {
        case ';':
            do
            {
                ch = is->GetCh(&eof);
                if (eof)
                    goto Eof;
            }
            while (ch != '\n' && ch != '\r');
            break;

        case '#':
            ch = is->GetCh(&eof);
            if (eof)
                ThrowError("Unexpected EOF", Undefined);

            switch (ch)
            {
            case 't':
            case 'T':
                return(True);

            case 'f':
            case 'F':
                return(False);

            case '\\':
                ch = is->GetCh(&eof);
                if (eof)
                    ThrowError("Unexpected EOF", Undefined);
                return(MakeCharacter(ch));

            case '(':
                return(ListToVector(this, ReadList(is, lnm)));
            }

            ThrowError("Unexpected character following #", Undefined);
            break;

        case '"':
            return(ReadString(is));

        case '(':
            if (lnm)
                return(lnm->Record(ReadList(is, lnm), is->GetLineNumber()));
            return(ReadList(is, lnm));

        case ')':
            if (rlf)
                return(EolObj);
            ThrowError("Unexpected ')'", Undefined);
            break;

        case '.':
            if (rlf)
                return(DotObj);
            ThrowError("Unexpected '.'", Undefined);
            break;

        case '\'':
            return(MakePair(SymbolQuote, MakePair(ReadInternal(is, lnm, 0, 0),
                    EmptyList)));

        default:
            if (SymbolChP(ch))
                return(ReadSymbolOrNumber(is, ch));
            break;
        }
    }

    return(0);

Eof:

    if (eaf == 0)
        ThrowError("Unexpected EOF", Undefined);
    return(EndOfFile);
}

TSObject * TSWorld::ReadString(TSInputStream * is)
{
    TSCh s[512];
    int sl = 0;
    TSCh ch;
    int eof;

    for (;;)
    {
        ch = is->GetCh(&eof);
        if (ch == '"')
            break;
        if (ch == '\\')
            ch = is->GetCh(&eof);
        if (eof)
            ThrowError("Unexpected EOF", Undefined);

        s[sl] = ch;
        sl += 1;
        if (sl == sizeof(s) / sizeof(TSCh))
            ThrowError("String too long", Undefined);
    }

    return(MakeString(sl, s));
}

TSObject * TSWorld::ReadSymbolOrNumber(TSInputStream * is, TSCh ch)
{
    TSCh s[128];
    int sl;
    int eof;
    TSFixnum n;
    int sdx;

    s[0] = ch;
    sl = 1;

    for (;;)
    {
        ch = is->GetCh(&eof);
        if (eof)
            ThrowError("Unexpected EOF", Undefined);

        if (SymbolChP(ch))
        {
            s[sl] = ch;
            sl += 1;
            if (sl == sizeof(s) / sizeof(TSCh))
                ThrowError("Symbol too long", Undefined);
        }
        else
        {
            is->UnGetCh(ch);
            break;
        }
    }

    if (StringAsNumber(sl, s, &n))
        return(MakeNumber(n));

    for (sdx = 0; sdx < sl; sdx++)
        s[sdx] = ChDownCase(s[sdx]);

    return(InternSymbol(sl, s));
}

TSObject * TSWorld::ReadList(TSInputStream * is, TSLineNumberMap * lnm)
{
    TSObject * obj;

    obj = ReadInternal(is, lnm, 0, 1);

    if (obj == EolObj)
        return(EmptyList);

    if (obj == DotObj)
    {
        obj = ReadInternal(is, lnm, 0, 0);
        if (ReadInternal(is, lnm, 0, 1) != EolObj)
            ThrowError("Bad dotted pair", Undefined);
        return(obj);
    }

    return(MakePair(obj, ReadList(is, lnm)));
}

void TSWorld::Write(TSOutputStream * os, TSObject * obj)
{
    if (obj == False)
        os->PutString("#f");
    else if (obj == True)
        os->PutString("#t");
    else if (obj == Unspecified)
        os->PutString("#<unspecified>");
    else if (obj == EmptyList)
        os->PutString("'()");
    else if (obj == EndOfFile)
        os->PutString("#<end-of-file>");
    else if (PairP(obj))
    {
        os->PutCh('(');
        for (;;)
        {
            Write(os, First(obj));
            if (PairP(Rest(obj)))
            {
                os->PutCh(' ');
                obj = Rest(obj);
            }
            else if (Rest(obj) == EmptyList)
            {
                os->PutCh(')');
                break;
            }
            else
            {
                os->PutString(" . ");
                Write(os, Rest(obj));
                os->PutCh(')');
                break;
            }
        }
    }
    else if (SymbolP(obj))
    {
        TSSymbol * sym = AsSymbol(obj);
        os->PutString(sym->Name->String, sym->Name->Length);
    }
    else if (StringP(obj))
    {
        TSString * s = AsString(obj);
        os->PutCh('"');
        os->PutString(s->String, s->Length);
        os->PutCh('"');
    }
    else if (CharacterP(obj))
    {
        os->PutString("#\\");
        os->PutCh(AsCharacter(obj));
    }
    else if (NumberP(obj))
    {
        TSCh s[64];
        int sl = NumberAsString(AsFixnum(obj), s, 10);
        os->PutString(s, sl);
    }
    else if (ProcedureP(obj))
    {
        TSCh s[64];
        int sl = NumberAsString((TSFixnum) obj, s, 16);
        os->PutString("#<procedure: #x");
        os->PutString(s, sl);
        os->PutString(">");
    }
    else if (BuiltinP(obj))
    {
        os->PutString("#<builtin: ");
        os->PutString(AsBuiltin(obj)->Info->Name);
        os->PutString(">");
    }
    else if (SyntaxP(obj))
    {
        os->PutString("#<syntax: ");
        if (AsSyntax(obj)->Name != 0)
            os->PutString(AsSyntax(obj)->Name);
        else
        {
            os->PutString("#x");
            TSCh s[64];
            int sl = NumberAsString((TSFixnum) obj, s, 16);
            os->PutString(s, sl);
        }
        os->PutString(">");
    }
    else if (VectorP(obj))
    {
        int vdx;

        os->PutString("#(");
        for (vdx = 0; vdx < AsVector(obj)->Length; vdx++)
        {
            if (vdx > 0)
                os->PutString(" ");
            Write(os, AsVector(obj)->Vector[vdx]);
        }
        os->PutString(")");
    }
    else if (Application->Write(os, obj) == 0)
    {
        TSCh s[64];
        int sl = NumberAsString((TSFixnum) obj, s, 16);
        os->PutString("#<unknown: #x");
        os->PutString(s, sl);
        os->PutString(">");
    }
}

void TSWorld::Display(TSOutputStream * os, TSObject * obj)
{
    TSCh ch;

    if (StringP(obj))
        os->PutString(AsString(obj)->String, AsString(obj)->Length);
    else if (CharacterP(obj))
    {
        ch = AsCharacter(obj);
        os->PutString(&ch, 1);
    }
    else if (Application->Display(os, obj) == 0)
        Write(os, obj);
}

static unsigned int Hash(int sl, TSCh * s)
{
    const char * p;
    unsigned int h = 0;

    sl *= sizeof(TSCh);
    for (p = (const char *) s; sl > 0; p++, sl--)
        h = (h << 5) - h + *p;

    return(h);
}

TSObject * TSWorld::InternSymbol(int sl, TSCh * s)
{
    int idx = Hash(sl, s) % (sizeof(Symbols) / sizeof(TSSymbol *));
    TSSymbol * sym = Symbols[idx];

    while (sym != 0)
    {
        if (sym->Name->Length == sl && StringEqual(s, sym->Name->String, sl))
            return(sym);

        sym = sym->Next;
    }

    sym = new TSSymbol(this, new TSString(this, sl, s), Symbols[idx]);
    Symbols[idx] = sym;

    return(sym);
}

TSObject * TSWorld::InternSymbol(char * s)
{
    TSCh ns[128];
    int nsl = 0;

    while (*s != 0)
    {
        ns[nsl] = *s;
        s += 1;
        nsl += 1;
        if (nsl == sizeof(ns) / sizeof(TSCh))
            ThrowError("Symbol too long", Undefined);
    }

    return(InternSymbol(nsl, ns));
}

TSObject * TSWorld::MakeString(int sl, TSCh * s)
{
    return(new TSString(this, sl, s));
}

TSObject * TSWorld::MakeVector(int vl, TSObject * obj)
{
    return(new TSVector(this, vl, obj));
}

TSObject * TSWorld::MakePair(TSObject * first, TSObject * rest)
{
    return(new TSPair(this, first, rest));
}

TSObject * TSWorld::MakeNumber(TSFixnum n)
{
    return((TSObject *) ((n << 2) | 0x1));
}

TSObject * TSWorld::MakeCharacter(TSCh ch)
{
    return((TSObject *) ((ch << 2) | 0x2));
}

TSObject * TSWorld::MakeProcedure(TSObject * name, TSObject * formals,
    TSObject * body, TSLineNumberMap * lnm)
{
    return(new TSProcedure(this, name, formals, body, lnm));
}

TSObject * TSWorld::MakeProcedure(TSObject * name, TSObject * formals,
    TSObject * body)
{
    return(MakeProcedure(name, formals, body, Stack->GetLineNumberMap()));
}

void TSWorld::AddBuiltin(TSBuiltinInfo * bi)
{
    AsSymbol(InternSymbol(bi->Name))->Value = new TSBuiltin(this, bi);
}

void TSWorld::AddSyntax(TSSyntaxInfo * si)
{
    AsSymbol(InternSymbol(si->Name))->Value = new TSSyntax(this, si->Name,
            si->Syntax);
}
